# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import uuid


import vulkn
from vulkn import Vulkn, session
from vulkn.utils import timer
from vulkn.types.base import TypeBase, func, Literal


# TODO: Urgent - Fix before release.


class ArrayVector(TypeBase):
    def __init__(self, value: any=None, name: str=None, n: str=None) -> None:
        self._sorted = False
        self._cache_table = None
        if isinstance(value, Literal):
            self._value = value
        else:
            v = []
            for col in value:
                if isinstance(col, str):
                    v.append("'{}'".format(col))
                else:
                    v.append(str(col))
            self._value = Literal("SELECT [{}] AS v".format(','.join(v)))

    # Creation

    @classmethod
    def range(self, start, end):
        if end - start > 100000000:
            raise Exception('ArrayVector cannot contain more than 100 million elements')
        q = f"""
            SELECT
                groupArray(number + {start}) AS v
            FROM numbers(toUInt64(1 + abs({start} - {end})))
            SETTINGS max_block_size = 100000000"""
        r = ArrayVector(Literal(q))
        r._sorted = True
        return r

    @classmethod
    def rand(self, start, end, length):
        if length > 100000000:
            raise Exception('ArrayVector cannot contain more than 100 million elements')
        q = f"""
            SELECT 
                groupArray({start} + rand64(number)%toUInt64(1 + abs({start} - {end}))) AS v
            FROM numbers({length})
            SETTINGS max_block_size = 100000000"""
        return ArrayVector(Literal(q))

    @classmethod
    def norm(self, mean, stddev, length):
        if length > 100000000:
            raise Exception('ArrayVector cannot contain more than 100 million elements')
        count = int(length)
        UInt32_MAX = vulkn.types.UInt32.MAX
        q = f"""
            SELECT
                arraySlice(
                    arrayReduce(
                        'groupArrayArray',
                        arrayMap(
                            i -> 
                                [ ((sqrt(-2.0*log(rand(i)/{UInt32_MAX}))*cos(2*pi()*rand(i+100000000)/{UInt32_MAX}))*toFloat32({stddev}))+toFloat32({mean})
                                , ((sqrt(-2.0*log(rand(i)/{UInt32_MAX}))*sin(2*pi()*rand(i+100000000)/{UInt32_MAX}))*toFloat32({stddev}))+toFloat32({mean})]
                            , range(toUInt64(ceil({count}/2))))), 1, {count}) AS v"""
        return ArrayVector(Literal(q))

    @classmethod
    def fromquery(self, query, column='v'):
        q = f"SELECT groupArray({column}) AS v FROM ({query}) SETTINGS max_block_size = 100000000"
        return ArrayVector(Literal(q))

    # Conversion

    def toColumnVector(self):
        from vulkn.types.column_vector import ColumnVector
        cache_table = self._cache()
        q = f"""
            SELECT i, _v AS v
            FROM (
                SELECT i, _v
                FROM ({cache_table})
                ARRAY JOIN
                    arrayEnumerate(v) AS i,
                    v AS _v)"""
        return ColumnVector(Literal(q))

    # Manage

    def alias(self, alias):
        self.vector_name = alias
        return self

    as_ = alias

    def _cache(self):
        ctx = Vulkn()
        return ctx.session.cache(self.sql, mode=session.CacheMode.Array)

    @timer
    def cache(self):
        cache_table = self._cache()
        r = ArrayVector(Literal("SELECT v FROM {}".format(cache_table)))
        r._sorted = self._sorted
        r._cache_table = cache_table
        return r

    def cast(self, to_type):
        toType = to_type if isinstance(to_type, str) else to_type.CAST
        q = "SELECT cast(v, 'Array({type})') AS v FROM ({vector})"
        r = ArrayVector(Literal(q.format(type=toType, vector=self.sql)))
        r._sorted = self._sorted
        return r

    @timer
    def exec(self):
        ctx = Vulkn()
        return ctx._conn.select(str(self.sql))

    @timer
    def show(self, count=10, max_block_size=30000):
        ctx = Vulkn()
        sql = self.sql
        q = f"""
            SELECT
                _i AS i,
                _v AS v
            FROM (
                SELECT
                    _i, _v
                FROM ({sql})
                ARRAY JOIN
                    arrayEnumerate(arraySlice(v, 1, {count})) AS _i,
                    arraySlice(v, 1, {count}) AS _v
            ) SETTINGS max_block_size={max_block_size}"""
        return ctx._conn.select(q).show()

    @timer
    def peek(self, count=10, max_block_size=30000):
        ctx = Vulkn()
        sql = self.sql
        q = f"""
            SELECT
                _i AS i,
                _v AS v
            FROM (
                SELECT
                    _i, _v
                FROM ({sql})
                ARRAY JOIN
                    arrayEnumerate(arraySlice(v, 1, {count})) AS _i,
                    arraySlice(v, 1, {count}) AS _v
            ) SETTINGS max_block_size={max_block_size}"""
        return ctx._conn.select(q).to_records()

    # Operations

    @timer
    def agg(self, agg):
        ctx = Vulkn()
        vector = self.sql
        q = f"SELECT arrayReduce('{agg}', v) AS v FROM ({vector})"
        return ctx._conn.select(q).to_records()

    def cut(self, cut_length):
        vector = self._value
        q = f"""
            SELECT
                groupArray(_v) AS v
            FROM (
                WITH
                    (_i-_i%{cut_length})/{cut_length}+1 AS _idx
                SELECT groupArray(_cut) AS _v
                FROM (
                    SELECT i - 1 AS _i, _cut FROM ({vector}) ARRAY JOIN v AS _cut, arrayEnumerate(v) AS i
                )
                GROUP BY _idx ORDER BY _idx
            ) SETTINGS max_block_size = 100000000"""
        r = ArrayVector(Literal(q))
        r._sorted = self._sorted
        return r

    def delta(self):
        vector = self._value
        q = f"""
            SELECT arrayConcat([NULL], groupArray(_delta)) AS v
            FROM (
                SELECT
                    v1 - v2 AS _delta
                FROM ({vector})
                ARRAY JOIN
                    v AS v1,
                    arrayConcat([NULL], arraySlice(v, 1, -1)) AS v2
            ) SETTINGS max_block_size = 100000000"""
        return ArrayVector(Literal(q))

    def flatten(self):
        return ArrayVector(
            Literal("SELECT groupArrayArrayArray(v) AS v FROM ({})".format(self._value)))

    @timer
    def index_agg(self, agg):
        ctx = Vulkn()
        vector = self.sql
        q = f"SELECT arrayReduce('{agg}', arrayEnumerate(v)) AS v FROM ({vector})"
        return ctx._conn.select(q).to_records()

    def join(self, N):
        vector = self._value
        other_vector = N._value
        q = f"""
            SELECT groupArray(_v) AS v
            FROM (
                SELECT _v FROM ({vector}) ARRAY JOIN v AS _v
                UNION ALL
                SELECT _v FROM ({other_vector}) ARRAY JOIN v AS _v
            ) SETTINGS max_block_size=100000000"""
        #q = "SELECT arrayConcat(({}),({})) AS v".format(self._value, N._value)
        return ArrayVector(Literal(q))

    def map(self, func, *args):
        a = ','.join(list(map(str, args))) + ',' if len(args) > 0 else ''
        vector = self.sql
        q = f"""
            SELECT groupArray(_v) AS v
            FROM (
                SELECT
                    {func}({a}_map) AS _v
                FROM ({vector})
                ARRAY JOIN
                    v AS _map
            ) SETTINGS max_block_size = 100000000"""
        return ArrayVector(Literal(q))

    def maplag(self, func):
        vector = self.sql
        q = f"""
            SELECT groupArray(_maplag) AS v
            FROM (
                SELECT
                    {func}(v2, v1) AS _maplag
                FROM ({vector})
                ARRAY JOIN
                    v AS v1,
                    arrayConcat([NULL], arraySlice(v, 1, -1)) AS v2
            ) SETTINGS max_block_size = 100000000"""
        return ArrayVector(Literal(q))

    def maplead(self, func):
        vector = self.sql
        q = f"""
            SELECT groupArray(_maplead) AS v
            FROM (
                SELECT
                    {func}(v2, v1) AS _maplead
                FROM ({vector})
                ARRAY JOIN
                    v AS v1,
                    arrayConcat(arraySlice(v, 2), [NULL]) AS v2
            ) SETTINGS max_block_size = 100000000"""
        return ArrayVector(Literal(q))

    def move(self, positions):
        vector = self._value
        q = None
        if positions > 0:
            q = f"""
                SELECT 
                    arrayConcat(
                        arrayWithConstant({positions}, NULL),
                        arraySlice(v, 1, -({positions}))
                    ) AS v
                FROM ({vector})"""
        elif positions < 0:
            q = f"""
                SELECT
                    arrayConcat(
                        arraySlice(v, abs({positions})+1),
                        arrayWithConstant(abs({positions}), NULL)
                    ) AS v
                FROM ({vector})"""
        else:
            return self
        r = ArrayVector(Literal(q))
        r._sorted = self._sorted
        return r

    def next(self):
        vector = self._value
        q = f"SELECT arrayConcat(arraySlice(v, 2), [NULL]) AS v FROM ({vector})"
        r = ArrayVector(Literal(q))
        r._sorted = self._sorted
        return r

    def prev(self):
        vector = self._value
        q = f"SELECT arrayConcat([NULL], arraySlice(v, 1, -1)) AS v FROM ({vector})"
        r = ArrayVector(Literal(q))
        r._sorted = self._sorted
        return r

    def shuffle(self):
        vector = self._value
        q = f"""
            SELECT
                groupArray(_shuffle) AS v
            FROM (
                SELECT _shuffle FROM (
                    SELECT _shuffle FROM ({vector}) ARRAY JOIN v AS _shuffle
                ) ORDER BY rand()
            ) SETTINGS max_block_size = 100000000"""
        return ArrayVector(Literal(q))

    def sort(self):
        if self._sorted:
            return self
        vector = self.sql
        q = f"""
            SELECT
                groupArray(_sort) AS v
            FROM (
                SELECT _sort FROM (
                    SELECT _sort FROM ({vector}) ARRAY JOIN v AS _sort
                ) ORDER BY _sort
            ) SETTINGS max_block_size = 100000000"""
        r = ArrayVector(Literal(q))
        r._sorted = True
        return r

    def sort_reverse(self):
        vector = self.sql
        q = f"SELECT arrayReverseSort(v) AS v FROM ({vector})"
        return ArrayVector(Literal(q))
  
    reverse = sort_reverse    

    def take(self, length):
        if length > 100000000:
            raise Exception('ArrayVector cannot contain more than 100 million elements')
        vector = self._value
        q = f"""
            SELECT
                _take AS v
            FROM (
                SELECT
                    groupArrayArrayArrayArray([
                        arrayMap(x -> v, range(toUInt64(floor({length} / length(v))))),
                        [arraySlice(v, 1, {length} % length(v))]
                    ]) AS _take
                FROM ({vector}))"""
        return ArrayVector(Literal(q))


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func
from vulkn.types import String, Array, Float64, UInt8, col, UInt64


def greatCircleDistance(lon1Deg, lat1Deg, lon2Deg, lat2Deg):
    return Float64(func('greatCircleDistance', lon1Deg, lat1Deg, lon2Deg, lat2Deg))

def pointInEllipses(*args):
    return UInt8(func('pointInEllipses', *args))

def pointInPolygon(point, vertices, polygons=None):
    if polygons:
        return UInt8(func('pointInPolygon', point, vertices, polygons))
    return UInt8(func('pointInPolygon', point, vertices))

def geohashEncode(longitude, latitude, precision=None):
    if precision:
        return String(func('geohashEncode', longitude, latitude, precision))
    return String(func('geohashEncode', longitude, latitude))

def geohashDecode(geohash):
    return col(func('geohashDecode', geohash))

def geoToH3(longitude, latitude, resolution):
    return UInt64(func('geoToH3', longitude, latitude, resolution))

def geohashesInBox(longitude_min, latitude_min, longitude_max, latitude_max, precision):
    return Array(
        func('geohashesInBox', longitude_min, latitude_min, longitude_max, latitude_max, precision))


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only

import subprocess
import os
import logging
import multiprocessing as mp
from functools import partial


from vulkn.clickhouse import sqlformat
from vulkn.utils import singleton, LogLevels, timer


log = logging.getLogger()


try:
    mp.set_start_method('forkserver')
except:
    pass


def CLIExecutor(log, cli, query):
    p = subprocess.run(cli, input=query, stdout=subprocess.PIPE, universal_newlines=True)
    return p.returncode


@singleton
class VulknScheduler:
    # TODO: This is a very naive scheduler used by the MergeVectors only. Needs far more work.
    def __init__(self, ctx, threads=None):
        self._ctx = ctx
        self._conn = self._ctx._conn
        self._total_threads = os.cpu_count()
        self._usable_threads = threads or int(self._total_threads*0.75)
        self._pool = None

    def dispatch(self, query):
        if isinstance(query, list):
            if len(query) == 1:
                query = query[0]
            else:
                return self._dispatch_pool(query)
        if isinstance(query, str):
            return self._dispatch_single(query)
        raise Exception('Unknown parameter to exec')

    def _dispatch_single(self, query):
        return self._conn.execute(query) == 0

    def _dispatch_pool(self, queries):
        flag_opts_overrides = {'disable_suggestion': True}
        opts_overrides = {'format': 'TSVWithNamesAndTypes'}
        cli = self._conn._cli(flag_opts_overrides=flag_opts_overrides, 
                              opts_overrides=opts_overrides,
                              settings_overrides=None)
        executor = partial(CLIExecutor, log, cli)
        if self._pool is None:
            self._pool = mp.Pool(self._usable_threads)
        [log.log(LogLevels.SQL, q) for q in queries]
        r = sum(self._pool.map(executor, queries))
        return r == 0


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func
from vulkn.types import UInt64, UInt8, Array, col


def bitmapBuild(array):
    return col(func('bitmapBuild', array))

def bitmapToArray(bitmap):
    return Array(func('bitmapToArray', bitmap))

def bitmapSubsetInRange(bitmap, range_start, range_end):
    return col(func('bitmapSubsetInRange', bitmap, range_start, range_end))

def bitmapContains(haystack, needle):
    return UInt8(func('bitmapContains', haystack, needle))

def bitmapHasAny(bitmap1, bitmap2):
    return UInt8(func('bitmapHasAny', bitmap1, bitmap2))

def bitmapHasAll(bitmap1, bitmap2):
    return UInt8(func('bitmapHasAll', bitmap1, bitmap2))

def bitmapAnd(bitmap1, bitmap2):
    return col(func('bitmapAnd', bitmap1, bitmap2))

def bitmapOr(bitmap1, bitmap2):
    return col(func('bitmapOr', bitmap1, bitmap2))

def bitmapXor(bitmap1, bitmap2):
    return col(func('bitmapXor', bitmap1, bitmap2))

def bitmapAndnot(bitmap1, bitmap2):
    return col(func('bitmapAndnot', bitmap1, bitmap2))

def bitmapCardinality(bitmap):
    return UInt64(func('bitmapCardinality', bitmap))

def bitmapAndCardinality(bitmap1, bitmap2):
    return UInt64(func('bitmapAndCardinality', bitmap1, bitmap2))

def bitmapOrCardinality(bitmap1, bitmap2):
    return UInt64(func('bitmapOrCardinality', bitmap1, bitmap2))

def bitmapXorCardinality(bitmap1, bitmap2):
    return UInt64(func('bitmapXorCardinality', bitmap1, bitmap2))

def bitmapAndnotCardinality(bitmap1, bitmap2):
    return UInt64(func('bitmapAndnotCardinality', bitmap1, bitmap2))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import os
import importlib
import inspect
import glob
from importlib import util


def load_library(ltype, basepath):
    modules = {}
    for f in filter(lambda x: not x.endswith('/__init__.py'), glob.glob(f'{basepath}/{ltype}/*.py')):
        module_name = f'{ltype}.' + os.path.basename(f)[0:-3]
        module_spec = importlib.util.spec_from_file_location(module_name, f)
        module = util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        for o in inspect.getmembers(module):
            if ((inspect.isfunction(o[1]) or inspect.isclass(o[1])) 
                    and o[1].__module__ == module_name and not o[0].startswith('_')):
                modules[o[0]] = o[1]
    return modules


def show():
    l = []
    for k in vectors.keys():
        l.append({'type': 'Vector Function', 'name': k})
    for k in extensions.keys():
        l.append({'type': 'SQL Extension', 'name': k})
    for k in udfs.keys():
        l.append({'type': 'User Defined Function', 'name': k})
    return l
    

vectors = load_library('vector', os.path.dirname(os.path.abspath(__file__)))
extensions = load_library('extensions', os.path.dirname(os.path.abspath(__file__)))
udfs = load_library('udfs', os.path.dirname(os.path.abspath(__file__)))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import TypeBase, func


class Numeric(TypeBase):
    _METHODS = {
        'exp': 'Float64',
        'log': 'Float64',
        'exp2': 'Float64',
        'log2': 'Float64',
        'exp10': 'Float64',
        'log10': 'Float64',
        'sqrt': 'Float64',
        'cbrt': 'Float64',
        'erf': 'Float64',
        'erfc': 'Float64',
        'lgamma': 'Float64',
        'tgamma': 'Float64',
        'sin': 'Float64',
        'cos': 'Float64',
        'tan': 'Float64',
        'asin': 'Float64',
        'acos': 'Float64',
        'atan': 'Float64',
        'roundToExp2': None,
        'roundDuration': None,
        'roundAge': None,
        'abs': None,
        'negate': None}

    def pow(self, right):
        return Float64(func('pow', self._value, right))

    def plus(self, right):
        return type(self)(func('plus', self._value, right))

    def minus(self, right):
        return type(self)(func('minus', self._value, right))

    def multiply(self, right):
        return type(self)(func('multiply', self._value, right))

    def divide(self, right):
        return type(self)(func('divide', self._value, right))

    def intDiv(self, right):
        return type(self)(func('intDiv', self._value, right))

    def intDivOrZero(self, right):
        return type(self)(func('intDivOrZero', self._value, right))

    def modulo(self, right):
        return type(self)(func('modulo', self._value, right))

    def gcd(self, right):
        return type(self)(func('gcd', self._value, right))

    def lcm(self, right):
        return type(self)(func('lcm', self._value, right))

    def floor(self, N: int=0):
        return type(self)(func('floor', self._value, N))
    
    def ceil(self, N: int=0):
        return type(self)(func('ceil', self._value, N))
    
    def round(self, N: int=0):
        return type(self)(func('round', self._value, N))

    def roundDown(self, arr):
        return type(self)(func('roundDown', self._value, arr))

    def __truediv__(self, right) -> any:
        return type(self)(func('divide', self._value, right))

    def __add__(self, right) -> any:
        return type(self)(func('plus', self._value, right))

    def __sub__(self, right) -> any:
        return type(self)(func('minus', self._value, right))

    def __mul__(self, right) -> any:
        return type(self)(func('multiply', self._value, right))

    def __floordiv__(self, right) -> any:
        return type(self)(self/right).floor()

    def __mod__(self, right) -> any:
        return type(self)(func('modulo', self._value, right))

    def __pow__(self, right) -> any:
        return type(self)(func('pow', self._value, right))

    def bitAnd(self, right) -> any:
        return type(self)(func('bitAnd', self._value, right))

    def bitOr(self, right) -> any:
        return type(self)(func('bitOr', self._value, right))
    
    def bitXor(self, right) -> any:
        return type(self)(func('bitXor', self._value, right))

    def bitNot(self) -> any:
        return type(self)(func('bitNot', self._value))

    def bitShiftLeft(self, right) -> any:
        return type(self)(func('bitShiftLeft', self._value, right))

    __lshift__ = bitShiftLeft

    def bitShiftRight(self, right) -> any:
        return type(self)(func('bitShiftRight', self._value, right))

    __rshift__ = bitShiftRight

    def bitRotateLeft(self, right) -> any:
        return type(self)(func('bitRotateLeft', self._value, right))

    def bitRotateRight(self, right) -> any:
        return type(self)(func('bitRotateRight', self._value, right))

    def bitTest(self, right) -> any:
        return type(self)(func('bitTest', self._value, right))

    def bitTestAll(self, right) -> any:
        return type(self)(func('bitTestAll', self._value, right))

    def bitTestAny(self, right) -> any:
        return type(self)(func('bitTestAny', self._value, right))


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.numeric import Numeric

class Float(Numeric):
    pass


class Float32(Float):
    TYPE = 'Float32'
    CAST = 'Float32'


class Float64(Float):
    TYPE = 'Float64'
    CAST = 'Float64'
    _METHODS = {
        'exp': 'Integer',
        'negate': 'Integer'}

    def e(self):
        return Float64(Literal('e()'))

    def pi(self):
        return Float64(Literal('pi()'))


from setuptools import setup, find_namespace_packages
from os import path


from vulkn.config import VERSION


BASE_URL = 'https://github.com/VulknData/vulkn'

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read().strip()

with open(path.join(this_directory, 'requirements.txt'), encoding='utf-8') as f:
    install_requires = f.read().strip().split('\n')

setup(
    name='vulkn',
    version=VERSION,
    description='The environmentally friendly petabyte scale Python eco-system built on Yandex ClickHouse',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='clickhouse',
    author='Jason Godden',
    author_email='jason@vulkndata.io',
    url=BASE_URL,
    download_url=f'{BASE_URL}/archive/v{VERSION}-alpha.tar.gz',
    license='gpl-3.0',
    include_package_data=True,
    packages=find_namespace_packages(include=['vulkn','vulkn.*']),
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Database',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    zip_safe=False,
    scripts=['scripts/vulkn'],
    python_requires='>=3.7'
)



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import copy
import logging
import random
import sys
from functools import partial


import vulkn.session
from vulkn.clickhouse import sqlformat
from vulkn.utils import VulknSQLFormatter, timer, LogLevels


log = logging.getLogger()


# TODO: Next release. Add table function datatables:
# https://clickhouse.yandex/docs/en/query_language/table_functions/


def copy_set(obj, key, *value):
    bool_keys = ['_distinct']
    this = copy.deepcopy(obj)
    if key in bool_keys:
        setattr(this, key, True)
    else:
        setattr(this, key, value)
    if key in this._op_chain:
        raise Exception(f'Operation {key} already set')
    else:
        this._op_chain.append(key)
    return this


class WriteMode:
    Replace = 0
    Append = 1
    Create = 2


class JoinStrictness:
    Any='ANY'
    All='ALL'
    AsOf='ASOF'
    Default=''


class JoinType:
    Left='LEFT'
    Right='RIGHT'
    LeftInner='LEFT INNER'
    RightInner='RIGHT INNER'
    Inner='INNER'
    Join=''
    LeftOuter='LEFT OUTER'
    RightOuter='RIGHT OUTER'
    Full='FULL'
    FullOuter='FULL OUTER'
    Cross='CROSS'


class VulknDataTable:   
    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == '_ctx':
                setattr(result, k, v)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result


def JoinDataTable(
        ctx, jointype, left, right, keys=None, strictness=JoinStrictness.All, global_mode=False):
    j = '({left}) {left_alias}{global_mode} {strictness} {join_type} JOIN ({right}) {right_alias}'
    left_alias = ''
    right_alias = ''
    if hasattr(left, '_alias') and left._alias is not None:
        left_alias = f'AS "{left._alias[0]}"'
    if hasattr(right, '_alias') and right._alias is not None:
        right_alias = f'AS "{right._alias[0]}"'
    q = j.format(left=left._get_query(), 
                 left_alias=left_alias,
                 global_mode='GLOBAL' if global_mode else '',
                 strictness=strictness,
                 join_type=jointype,
                 right=right._get_query(),
                 right_alias=right_alias)
    left_cols = [f'{l._name} AS "{left._alias[0]}.{l._name}"' 
        if left_alias else l for l in left.get_schema()]
    right_cols = [f'{r._name} AS "{right._alias[0]}.{r._name}"'
        if right_alias else r for r in right.get_schema()]
    if keys:
        q = '{} USING ({})'.format(q, ','.join(keys))
    return SelectQueryDataTable(ctx, q).select(*[*left_cols, *right_cols])


def NumbersDataTable(ctx, count, system=False, mt=False):
    if system:
        t = 'system.numbers' + ('_mt' if mt else '')
        return SelectQueryDataTable(ctx, t).limit(count).select('*')
    else:
        t = 'numbers' + ('_mt' if mt else '')
        return SelectQueryDataTable(ctx, '{}({})'.format(t, count)).select('*')


def RandomDataTable(ctx, count, start=0, end=((sys.maxsize*2)+1), system=False, mt=False):
    q = f'{start} + rand()%(if(mod == 0, mod - 1, ({end}-({start})+1) AS mod)) AS number'
    if system:
        t = 'system.numbers' + ('_mt' if mt else '')
        return SelectQueryDataTable(ctx, t).limit(count).select(q)
    else:
        t = 'numbers' + ('_mt' if mt else '')
        return SelectQueryDataTable(ctx, '{}({})'.format(t, count)).select(q)


def RandomFloatDataTable(ctx, count, start=0, end=((sys.maxsize*2)+1), system=False, mt=False):
    maxsize = ((sys.maxsize*2)+1)
    q = f"{start} + rand64()%({end}-({start})) + rand64()/{maxsize} AS number"
    if system:
        t = 'system.numbers' + ('_mt' if mt else '')
        return SelectQueryDataTable(ctx, t).limit(count).select(q)
    else:
        t = 'numbers' + ('_mt' if mt else '')
        return SelectQueryDataTable(ctx, '{}({})'.format(t, count)).select(q)


def OneDataTable(ctx):
    return SelectQueryDataTable(ctx, 'system.one').select('*')


def RangeDataTable(ctx, start, end, system=False, mt=False):
    if end <= start:
        raise Exception('end must be greater than start')
    if system or mt:
        t = 'system.numbers' + ('_mt' if mt else '')
        return (SelectQueryDataTable(ctx, t)
                .limit(end + 1 -(start))
                .select('number + {} AS range'.format(str(start))))
    else:
        return (SelectQueryDataTable(ctx, 'numbers({})'.format(end + 1 -(start)))
                .select('number + {} AS range'.format(str(start))))


def aj(datatables, keys, global_mode=False):
    df = datatables[0]
    for d in datatables:
        df = df.aj(d, keys, global_mode)
    return df


class ShowSQLMixin:
    def _show_sql(self, **params):
        q = self._get_query()
        sql = VulknSQLFormatter().format(q, **params) if params else q
        return sql
 
    def show_sql(self, **params):
        q = self._get_query()
        sql = VulknSQLFormatter().format(q, **params) if params else q
        statements = vulkn.sql.SQLMessage(sql).statements()
        if len(statements) > 1:
            raise Exception('More than one statement is not supported')
        return sqlformat(statements[0].optimize())


class CachedQueryMixin:
    def cache(self, engine=None):
        if engine is None:
            engine = vulkn.engines.Memory()
        q = self._get_query()
        sql = VulknSQLFormatter().format(q, **self._params[0]) if self._params else q
        statements = vulkn.sql.SQLMessage(sql).statements()
        if len(statements) > 1:
            raise Exception('More than one statement is not supported')
        t = self._ctx.session.cache(statements[0].optimize(), engine=engine)
        return SelectQueryDataTable(self._ctx, t).select('*')


class TableWriterMixin:
    def write(self, database, table, engine=None, mode=WriteMode.Create):
        table_exists = self._ctx.tableExists(database, table)
        if table_exists and mode == WriteMode.Create:
            raise Exception('Table exists and invalid write mode specified')
        if engine is None:
            engine = vulkn.engines.Memory()
        if mode != WriteMode.Append:
            self._ctx.dropTable(database, table)
        q = self._get_query()
        source = VulknSQLFormatter().format(q, **(self._params)) if self._params else q
        statements = vulkn.sql.SQLMessage(source).statements()
        if len(statements) > 1:
            raise Exception('More than one statement is not supported')
        sql = statements[0].optimize()
        if mode != WriteMode.Append:
            sql = f'CREATE TABLE {database}.{table} ENGINE = {engine} AS {sql}'
        else:
            sql = f'INSERT INTO {database}.{table} {sql}'
        if not self._ctx.exec(sql):
            raise Exception('Unable to write table')
        return SelectQueryDataTable(self._ctx, f'{database}.{table}').select('*')


class QueryExecutorMixin:
    def plan(self, **params):
        q = self._get_query()
        statements = vulkn.sql.SQLMessage(VulknSQLFormatter().format(q, **params)).statements()
        if len(statements) > 1:
            raise Exception('More than one statement is not supported')
        return QueryStringDataTable(self._ctx, statements[0].optimize()).select('*')

    @timer
    def exec(self, **params):
        q = self._get_query()
        statements = vulkn.sql.SQLMessage(VulknSQLFormatter().format(q, **params)).statements()
        if len(statements) > 1:
            raise Exception('More than one statement is not supported')
        return self._ctx._conn.select(statements[0].optimize())

    def get_schema(self):
        name = f'v{random.randint(1,sys.maxsize)}'
        c = None
        try:
            self._ctx.drop_view('vulkn', name)
            self._ctx.create_view('vulkn', name, self.show_sql()[0:-1])
            c = self._ctx.get_schema('vulkn', name)
        finally:
            self._ctx.drop_view('vulkn', name)
        return c

    def to_records(self):
        return self.exec().to_records()

    def to_pandas(self):
        return self.exec().to_pandas()

    def show(self):
        return self.exec().show()

    def to_list(self):
        return self.exec().to_list()

    def to_dict(self):
        return self.exec().to_dict()

    def to_recarray(self):
        return self.exec().to_recarray()

    r = property(to_records)
    s = property(show)
    p = property(to_pandas)
    l = property(to_list)
    d = property(to_dict)
    n = property(to_recarray)
    e = exec


class BaseTableDataTable:
    def __init__(self, ctx, database, table):
        self._ctx = ctx
        self._database = database
        self._table = table

    def __str__(self):
        return '"{database}"."{table}"'.format(database=self._database, table=self._table)

    def __getattr__(self, attrname):
        return getattr(SelectQueryDataTable(self._ctx, str(self)), attrname)

    def desc(self):
        return self._ctx.q(f'DESC {self._database}.{self._table}').exec()

    def drop_columns(self, *columns):
        return self._ctx.drop_columns(self._database, self._table, columns)

    def load(self, parts=None):
        from vulkn.types import c
        dt = SelectQueryDataTable(self._ctx, str(self)).all()
        if parts:
            dt = dt.where(c('_part').in_(parts))
        q = dt._get_query()
        sql = VulknSQLFormatter().format(q, **self._params[0]) if self._params else q
        statements = vulkn.sql.SQLMessage(sql).statements()
        if len(statements) > 1:
            raise Exception('More than one statement is not supported')
        t = self._ctx.session.cache(statements[0].optimize(), engine=vulkn.engines.Memory())
        return BaseTableDataTable(self._ctx, *(t.split('.')))


class QueryStringDataTable(VulknDataTable, QueryExecutorMixin, ShowSQLMixin, CachedQueryMixin, TableWriterMixin):
    def __init__(self, ctx, query):
        self._ctx = ctx
        self._query = query

    def _get_query(self):
        return self._query

    def __getattr__(self, attrname):
        attrs = ['limit','where','filter','limit_by','limitBy','head','first','distinct',
            'order_by','orderBy','sort','prewhere','preWhere']
        this = SelectQueryDataTable(self._ctx, '({})'.format(self._get_query()))
        if attrname.lower() in attrs:
            return getattr(this.select('*'), attrname)
        else:
            return getattr(this, attrname)


class SelectQueryDataTable(VulknDataTable,
                           QueryExecutorMixin,
                           ShowSQLMixin,
                           CachedQueryMixin,
                           TableWriterMixin):
    def __init__(self, ctx, source=None):
        self.max_rows = 100
        self.max_preview_rows = 30
        self._ctx = ctx
        self._columns = None
        self._distinct = False
        self._table = (source,) if source is not None else None
        self._prewhere = None
        self._where = None
        self._group_by = None
        self._group_by_with_totals = False
        self._order_by = None
        self._having = None
        self._limit = None
        self._limit_by = None
        self._vectorize_by = None
        self._join = None
        self._with_ = None
        self._params = None
        self._array_join = None
        self._left_array_join = None
        self._alias = None
        self._op_chain = []

    def __getattr__(self, attrname):
        attrs = {
            'alias': None,
            'as_': '_alias',
            'with_': None,
            'distinct': None,
            'from_': '_table',
            'prewhere': None,
            'preWhere': '_prewhere',
            'where': None,
            'filter': '_where',
            'order_by': None,
            'orderBy': '_order_by',
            'sort': '_order_by',
            'having': None,
            'limit': None,
            'params': None,
            'array_join': None,
            'arrayJoin': '_array_join',
            'left_array_join': None,
            'leftArrayJoin': '_left_array_join'
        }

        if attrname in attrs.keys():
            attr = attrs[attrname] or '_{}'.format(attrname.lower())
            dt = self._subquery() if getattr(self, attr) is not None else self
            return partial(copy_set, dt, attr)
        else:
            raise AttributeError

    def _subquery(self):
        return SelectQueryDataTable(self._ctx, self)

    def all(self):
        cols = ('*',)
        if not isinstance(self._table[0], type(self)):
            try:
                table = self._table[0].rsplit('.')[1].strip('"')
                cols = [t._name for t in self._ctx.getSchema(table)]
            except:
                cols = ('*',)
        return self.select(*cols)

    def select(self, *cols):
        if not cols:
            cols = ('*',)
        if self._columns:
            return self._subquery().select(*cols)
        else:
            return copy_set(self, '_columns', *cols)
   
    def count(self):
        return self._subquery().select('count()')
 
    def first(self):
        return self.head(1)

    def head(self, rows=1):
        return copy_set(self, '_limit', rows)

    def limit_by(self, limit, by):
        dt = self._subquery() if self._limit_by else self
        return copy_set(dt, '_limit_by', limit, by)

    limitBy = limit_by

    def group_by(self, *cols, with_totals=False):
        r = None
        if self._group_by:
            r = self._subquery().group_by(*cols)
        else:
            r = copy_set(self, '_group_by', *cols)
        r._group_by_with_totals = with_totals
        return r

    groupBy = group_by

    def vectorize_by(self, key, sort, attributes=None):
        if self._vectorize_by or self._group_by or self._order_by:
            return self._subquery().all().vectorize_by(key, sort, attributes)
        else:
            return copy_set(self, '_vectorize_by', key, attributes, sort)

    vectorizeBy = vectorize_by

    def join(self, jointype, right, keys=None, strictness=JoinStrictness.All, global_mode=False):
        return JoinDataTable(self._ctx, jointype, self, right, keys, strictness, global_mode)

    def ej(self, right, keys, strictness=JoinStrictness.All, global_mode=False):
        return self.join(JoinType.Inner, right, keys, strictness, global_mode)

    def rj(self, left, keys, strictness=JoinStrictness.All, global_mode=False):
        return self.join(JoinType.Right, left, keys, strictness, global_mode)

    def lj(self, right, keys, strictness=JoinStrictness.All, global_mode=False):
        return self.join(JoinType.Left, right, keys, strictness, global_mode)

    def fj(self, right, keys, strictness=JoinStrictness.All, global_mode=False):
        return self.join(JoinType.Full, right, keys, strictness, global_mode)

    def aj(self, right, keys, global_mode=False):
        return self.join(JoinType.Join, right, keys, JoinStrictness.AsOf, global_mode)

    def cj(self, right, global_mode=False):
        return self.join(JoinType.Cross,
                         right,
                         keys=None,
                         strictness=JoinStrictness.Default,
                         global_mode=global_mode)

    def _get_query(self):
        q = 'SELECT'
        if self._with_:
            q = 'WITH {} {}'.format(', '.join(map(str, self._with_)), q)
        if self._distinct:
            q = f'{q} DISTINCT'
        if self._columns:
            q = '{} {}'.format(q, ', '.join(map(str, self._columns)))
        else:
            q = f'{q} *'
        if self._join:
            q = '{} FROM {}'.format(q, str(self._join))
        elif self._table:
            if isinstance(self._table[0], type(self)):
                q = '{} FROM ({})'.format(q, self._table[0]._get_query())
                if hasattr(self._table[0], '_alias') and self._table[0]._alias is not None:
                    q = '{} AS "{}"'.format(q, self._table[0]._alias[0])
            else:
                q = '{} FROM {}'.format(q, ', '.join(map(str, self._table)))
        if self._array_join or self._left_array_join:
            if self._array_join:
                q = '{} ARRAY JOIN {}'.format(q, ', '.join(map(str, self._array_join)))
            else:
                q = '{} LEFT ARRAY JOIN {}'.format(q, ', '.join(map(str, self._left_array_join)))
        if self._prewhere:
            q = '{} PREWHERE {}'.format(q, ' AND '.join(map(str, self._prewhere)))
        if self._where:
            q = '{} WHERE {}'.format(q, ' AND '.join(map(str, self._where)))
        if self._vectorize_by:
            if self._vectorize_by[1]:
                q = '{} VECTORIZE BY ({}, {}, {})'.format(q, *self._vectorize_by)
            else:
                q = '{} VECTORIZE BY ({}, {})'.format(q,
                                                      self._vectorize_by[0],
                                                      self._vectorize_by[2])
        if self._group_by:
            q = '{} GROUP BY {}'.format(q, ', '.join(map(str, self._group_by)))
            if self._group_by_with_totals:
                q = f'{q} WITH TOTALS'
            if self._having:
                q = '{} HAVING {}'.format(q, ', '.join(map(str, self._having)))
        if self._order_by:
            q = '{} ORDER BY {}'.format(q, ', '.join(map(str, self._order_by)))
        if self._limit_by:
            q = '{} LIMIT {} BY {}'.format(q, 
                                           str(self._limit_by[0]),
                                           ', '.join(map(str, self._limit_by[1])))
        if self._limit:
            q = '{} LIMIT {}'.format(q, str(self._limit[0]))
            if len(self._limit) == 2:
                q = '{} OFFSET {}'.format(q, str(self._limit[1]))
        return q

    #def __repr__(self):
    #    return str(self.limit(self.max_rows).exec().show_pandas(self.max_preview_rows))


class UpdateQueryDataTable(VulknDataTable, ShowSQLMixin):
    def __init__(self, ctx, table):
        self._ctx = ctx
        self._table = table
        self._updates = None
        self._where = None
    
    def __getattr__(self, attrname):
        attrs = { 'set_': '_updates', 'where': None }
        if attrname in attrs.keys():
            attr = attrs[attrname] or '_{}'.format(attrname.lower())
            return partial(copy_set, self, attr)
        else:
            raise AttributeError

    def sync(self):
        self._sync = True
        return self

    def _get_query(self):
        where = ' AND '.join(self._where or ['1 = 1'])
        columns = ', '.join(self._updates)
        table = self._table
        return f'ALTER TABLE {table} UPDATE {columns} WHERE {where}'

    def exec(self):
        r = self._ctx.exec(self._get_query())
        d = self._ctx.table('system.mutations').select('*').where("database || '.' || table = '{}'".format(self._table)).exec()
        print(d)
        if not self._sync:
            return r
        print(r)
        


class DeleteQueryDataTable(VulknDataTable, ShowSQLMixin):
    def __init__(self, ctx, table):
        self._ctx = ctx
        self._table = table
        self._where = None

    def __getattr__(self, attrname):
        attrs = {'where': None}
        if attrname in attrs.keys():
            attr = attrs[attrname] or '_{}'.format(attrname.lower())
            return partial(copy_set, self, attr)
        else:
            raise AttributeError

    def _get_query(self):
        table = self._table
        if self._where is None:
            return f'TRUNCATE TABLE {table}'
        where = ' AND '.join(self._where)
        return f'ALTER TABLE {table} DELETE WHERE {where}'

    def sync(self):
        self._sync = True
        return self

    def exec(self):
        mutations = (self._ctx
                     .table('system.mutations')
                     .select('*')
                     .where("database || '.' || table = '{}'".format(self._table),
                            'is_done = 0',
                            "latest_fail_reason != ''",
                            "command LIKE 'DELETE%'")
                     .exec()
                     .to_records())
        if len(mutations) > 0:
            errors = [d['mutation_id'] + ': ' + d['latest_fail_reason'] for d in mutations]
            raise Exception('There are unfinished failed mutations for this table: \n' + '\n'.join(errors))
        r = self._ctx.exec(self._get_query())
        d = self._ctx.table('system.mutations').select('*').where("database || '.' || table = '{}'".format(self._table)).exec().to_records()
        if len(d[0]['latest_fail_reason']) > 0:
            raise Exception('Delete operation failed: ' + d[0]['latest_fail_reason'])
        if not self._sync:
            return r
        return r



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import sys
import random


from vulkn.sql.utils import ast_protect_injection


@ast_protect_injection
def vectorRowNumber():
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}'
    arr_join_expr = (f'arrayEnumerate(`--#v`.1) AS v{varg}',)
    return (select_expr, arr_join_expr)

vectorRowNum = vectorRowNumber

@ast_protect_injection
def vectorAgg(agg, v):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}'
    arr_join_expr = (f"arrayWithConstant(length(__{v}), arrayReduce('{agg}', __{v})) AS v{varg}",)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorFirst(v):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}'
    arr_join_expr = (f"arrayWithConstant(length(__{v}), (__{v})[1]) AS v{varg}",)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorLast(v):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}'
    arr_join_expr = (f"arrayWithConstant(length(__{v}), (__{v})[-1]) AS v{varg}",)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorLag(v, offset=-1):
    offset = int(str(offset))
    if offset > -1:
        raise Exception('Parameter "offset" must be less than 0')
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}'
    arr_join_expr = (f"arrayConcat(arrayWithConstant(abs({offset}), NULL), arraySlice(__{v}, 1, {offset})) AS v{varg}",)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorLead(v, offset=1):
    offset = int(str(offset))
    if offset < 1:
        raise Exception('Parameter "offset" must be greater than 0')
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}'
    arr_join_expr = (f"arrayConcat(arraySlice(__{v}, 1+{offset}), arrayWithConstant({offset}, NULL)) AS v{varg}",)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorDenseRank(v):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}'
    arr_join_expr = (f'arrayEnumerateDense(__{v}) AS v{varg}',)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorCumeDist(v):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}_1 / v{varg}_2'
    arr_join_expr = (f'arrayEnumerate(__{v}) AS v{varg}_1',
                     f'arrayWithConstant(length(__{v}), length(__{v})) AS v{varg}_2',)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorPercentRank(v):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}_1 / v{varg}_2'
    arr_join_expr = (f'arrayConcat([0], arrayEnumerate(arraySlice(__{v}, 2))) AS v{varg}_1',
                     f'arrayWithConstant(length(__{v}), length(__{v}) - 1) AS v{varg}_2',)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorMap(func, v):
    varg = random.randint(1,sys.maxsize)
    func = str(func).strip("'")
    select_expr = f'{func}(v{varg})'
    arr_join_expr = (f'__{v} AS v{varg}',)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorNth(v, offset=1):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}'
    arr_join_expr = (f'arrayWithConstant(length(__{v}), (__{v})[{offset}]) AS v{varg}',)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorNtile(v, buckets):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'ceil(v{varg}_2 / v{varg}_1)'
    arr_join_expr = (f'arrayWithConstant(length(__{v}), (length(__{v})/{buckets})) AS v{varg}_1',
                     f'arrayEnumerate(__{v}) AS v{varg}_2',)
    return (select_expr, arr_join_expr)

@ast_protect_injection
def vectorRank(v):
    varg = random.randint(1,sys.maxsize)
    select_expr = f'v{varg}_2 - v{varg}_1 + 1'
    arr_join_expr = (f'arrayEnumerateUniq(__{v}) AS v{varg}_1',
                     f'arrayEnumerate(__{v}) AS v{varg}_2',)
    return (select_expr, arr_join_expr)



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import itertools
import logging
import textwrap


import vulkn.sql.utils as utils
import vulkn.sql.library as vlib
from vulkn.contrib.sqlparse import sqlparse


log = logging.getLogger()
__priority__ = 10


def ASTRewriteVectorizeByClause(ast):
    ret = sqlparse.parse(str(list(_AnalyzeVectorizeByAST(ast))[0]))[0]
    return ret


def _AnalyzeVectorizeByAST(ast):
    original_tree = ast.tokens if isinstance(ast, sqlparse.sql.TokenList) else ast
    current_tree = []
    subquery_found = False
    vectorize_found_idx = 0
    for idx, token in enumerate(original_tree):
        if subquery_found and isinstance(token, sqlparse.sql.Parenthesis):
            t = sqlparse.sql.TokenList(
                list(_AnalyzeVectorizeByAST(sqlparse.sql.TokenList(token.tokens[1:-1]))))
            current_tree.append(sqlparse.sql.Token('Literal', f' ({t}) '))
            subquery_found = False
            continue
        elif token.ttype == sqlparse.tokens.Keyword:
            if token.normalized == 'FROM' or 'JOIN' in token.normalized:
                if isinstance(ast.token_next(idx)[1], sqlparse.sql.Parenthesis):
                    subquery_found = True
            elif token.value.upper().startswith('VECTORIZE'):
                vectorize_found_idx = idx
        current_tree.append(token)
    a = sqlparse.sql.TokenList(current_tree)
    if vectorize_found_idx > 0:
        yield _VectorizeByASTRewriter(a, vectorize_found_idx).rewrite()
    else:
        yield a


def _RewriteVectorFunctionAST(ast):
    if isinstance(ast, sqlparse.sql.Function):
        if ast.tokens[0].value in vlib.vectors.keys():
            parameters = _RewriteVectorFunctionAST(ast.get_parameters())
            yield vlib.vectors[ast.tokens[0].value](*list(parameters))                
        else:
            yield token
    tree = ast.tokens if isinstance(ast, sqlparse.sql.TokenList) else ast
    for idx, token in enumerate(tree):
        if isinstance(token, sqlparse.sql.Parenthesis):
            if len(token.tokens) > 1 or not isinstance(token.tokens[0], sqlparse.sql.Token):
                for t in _RewriteVectorFunctionAST(token):
                    yield t
            else:
                yield token
        elif isinstance(token, sqlparse.sql.Function):
            if token.tokens[0].value in vlib.vectors.keys():
                parameters = _RewriteVectorFunctionAST(token.get_parameters())
                yield vlib.vectors[token.tokens[0].value](*list(parameters))                
            else:
                yield token
        elif isinstance(token, sqlparse.sql.Identifier):
            if len(token.tokens) > 1 or not isinstance(token.tokens[0], sqlparse.sql.Token):
                for t in _RewriteVectorFunctionAST(token):
                    yield t
            else:
                yield token
        elif isinstance(token, sqlparse.sql.IdentifierList):
            for t in _RewriteVectorFunctionAST(token):
                yield t
        else:
            yield token


class _VectorizeByASTRewriter():
    def __init__(self, ast, curr_idx):
        self._tokens = ast
        self._ast = ast
        self._curr_idx = curr_idx

    def rewrite(self):
        self.validate()
        p = self._parse()
        q = self._compile(
            p['from_clause'],
            p['where_clause'],
            p['key'],
            p['sort'],
            p['cols'],
            p['chunk_size'],
            p['metrics'],
            p['vectors'],
            p['expr']
        )
        return sqlparse.parse(q)[0]

    def validate(self):
        excludes = ['GROUP BY', 'ORDER BY', 'LIMIT BY', 'LIMIT']
        if not utils.validate_statement_excludes(excludes, self._ast):
            raise Exception('GROUP BY, ORDER BY, LIMIT BY and LIMIT are invalid in VECTORIZE BY statements')

    def _parse(self):
        def _parse_identifiers(ids, key, sort, cols):
            metrics = []
            vectors = []
            expr = []
            for id in utils.resolve_identifiers(ids):
                if len(set(vlib.vectors.keys()).intersection(set(id['funcs']))) > 0:
                    vectors.append(id)
                elif id['id'] != key and id['id'] != sort:
                    expr.append(id)
                for src in id['src']:
                    if not (src == key or src == sort or src in cols):
                        metrics.append(id)
            return (metrics, vectors, expr)
    
        args = utils.get_keyword_function_args(self._ast, self._curr_idx)
        # TODO: Chunking. Default to 1.
        # (key, sort, cols) = (args[0], args[-2], args[1:-2] if len(args) > 3 else [])
        (key, sort, cols) = (args[0], args[-1], args[1:-1] if len(args) > 3 else [])
        (metrics, vectors, expr) = _parse_identifiers(utils.ast_column_projections(self._ast), key, sort, cols)

        r = {    
            'from_clause': utils.ast_from_clause(self._ast),
            'where_clause': utils.ast_where_clause(self._ast),
            'key': key,
            'sort': sort,
            'cols': cols,
            # TODO: Chunking. Default to 1.
            #'chunk_size': int(args[-1]) if args[-1].isnumeric() else 2,
            'chunk_size': 1,
            'metrics': metrics,
            'vectors': vectors,
            'expr': expr
        }

        return r

    def _compile(self, from_clause, where_clause, key, sort, cols, chunk_size, metrics, vectors, expr):
        vector_funcs = [list(_RewriteVectorFunctionAST(f['node'])) for f in vectors]
        expr_cols = list([str(f['node']) for f in expr])
        metric_ids = list(set(list(itertools.chain.from_iterable([m['src'] for m in metrics]))))
        metric_map = dict((metric, idx+2) for idx, metric in enumerate(metric_ids))
        w = where_clause if where_clause is not None else ''
        # TODO: String replacement is hacky.
        aj_metric_vectors = list(itertools.chain.from_iterable(list(vf)[0][1] for vf in vector_funcs))
        for i, _ in enumerate(aj_metric_vectors):
            for k, v in metric_map.items():
                aj_metric_vectors[i] = aj_metric_vectors[i].replace(f'__{k}', f'`--#v`.{v}')
        metric_vector_aliases = itertools.chain(
            ['`--#v`.1 AS `{0}`'.format(sort)],
            [f'`--#v`.{v} AS `{k}`' for k, v in metric_map.items()],
            aj_metric_vectors)
        vector_aggs = [str(vector_funcs[i][0][0]) + ' AS `' + vectors[i]['id'] + '`' for i in range(len(vector_funcs))]
        query_template = textwrap.dedent("""\
            SELECT
                {final_projection},
                {sort},
                {expr}
                {vector_aggs}
            FROM
            (
                SELECT
                    {initial_projection},
                    arraySort(x -> x.1, groupArray(({sort}, {vector_group}))) AS `--#v`
                FROM {initial_from} {where_clause}
                GROUP BY {initial_projection}
            )
            ARRAY JOIN
                {metric_vector_aliases}
            SETTINGS enable_unaligned_array_join = 1""")
        q = query_template.format(
            sort=sort,
            initial_projection=', '.join(itertools.chain([key], cols)),
            initial_from=from_clause,
            where_clause=w,
            vector_group=', '.join(metric_ids),
            vector_cols=(',\n'+' '*8).join([f'`--#v`.{v} AS `__{k}`' for k, v in metric_map.items()]),
            chunk_size=chunk_size,
            metric_vector_aliases=(',\n' + ' '*4).join(metric_vector_aliases),
            vector_aggs=(',\n'+' '*4).join(vector_aggs),
            final_projection=', '.join(itertools.chain([key], cols)),
            expr=(', '.join(expr_cols) + ',' if len(expr_cols) > 0 else ''))
        return q




# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import random
import math
import hashlib
import os


import vulkn
from vulkn.types.base import func, Literal
from vulkn.utils import timer


class MergeVector:
    def __init__(self, max_block_size=10000, max_threads=2, array_chunk_size=100000000, uuid=None):
        if array_chunk_size > 100000000:
            raise Exception('array_chunk_size cannot be greater than 100 million')
        self._ref = None
        self._arrays = None
        self._max_block_size = max_block_size
        self._max_threads = max_threads
        self._array_chunk_size = array_chunk_size
        self._uuid = uuid or MergeVector.generateUUID()
        self._length = None
        self._type = 'UInt64'
        self._sorted = False

    def toArrayVector(self):
        from vulkn.types.array_vector import ArrayVector
        return ArrayVector(
            Literal(
                "SELECT v FROM {} WHERE i < 100000000 ORDER BY _table, i LIMIT 100000000".format(
                    self._ref)))

    def _session(self):
        return vulkn.Vulkn().session.session_store

    def _rebuild_arrays(self):
        ctx = vulkn.Vulkn()
        rename = [f'RENAME TABLE {chunk}_ TO {chunk}' for chunk in self._arrays]
        drop = [f'DROP TABLE IF EXISTS {chunk}' for chunk in self._arrays]
        ctx.scheduler.dispatch(drop)
        ctx.scheduler.dispatch(rename)

    def _rebuild_ref(self):
        ctx = vulkn.Vulkn()
        parent = "CREATE TABLE {ref} AS {ref}_1 ENGINE = Merge('{vulkn_database}', '^{session}_{uuid}_([0-9]*)$')"
        ctx.scheduler.dispatch('DROP TABLE IF EXISTS {ref}'.format(ref=self._ref))
        ctx.scheduler.dispatch(parent.format(ref=self._ref, 
                                             vulkn_database=self._session().split('.')[0],
                                             session=self._session().split('.')[1],
                                             uuid=self._uuid))

    @staticmethod
    def generateUUID():
        return hashlib.md5(str(random.random()).encode()).hexdigest()

    def chunks(self, length, chunk_size):
        r = [math.floor(length/chunk_size), length%chunk_size]
        if r[0] == 1 and r[1] == 0 and self._length is not None:
            r[1] = length/self._length
        return r

    def __getattr__(self, attrname):
        attrs = [
            'count', 'min','max','median','avg','sum','any','anyHeavy','anyLast','skewPop','skewSamp',
            'kurtPop','kurtSamp','uniq','uniqCombined','uniqExact','varSamp','varPop','stddevSamp',
            'stddevPop']
        if attrname in attrs:
            return lambda: self.agg(attrname)
        else:
            raise AttributeError

    @classmethod
    @timer
    #def rand(self, min_value, max_value, length, max_block_size=10000, max_threads=2, array_chunk_size=5000000):
    def rand(self, min_value, max_value, length, max_block_size=None, max_threads=2, array_chunk_size=None):
        sql = 'CREATE TABLE {session}_{uuid}_{chunk} ENGINE=Memory() AS '
        sql += 'SELECT number AS i, rand64()%toUInt64({mod}) AS v FROM numbers_mt({array_chunk_size}) '
        sql += 'SETTINGS max_block_size={max_block_size}, max_threads={max_threads}'
        array_chunk_size = array_chunk_size or math.ceil(length/((os.cpu_count()*0.75)-2))
        if array_chunk_size > 100000000:
            array_chunk_size = 100000000
        max_block_size = max_block_size or array_chunk_size
        uuid = MergeVector.generateUUID()
        r = MergeVector(uuid=uuid,
                        max_block_size=max_block_size,
                        max_threads=max_threads,
                        array_chunk_size=array_chunk_size)
        r._ref = '{session}_{uuid}'.format(session=r._session(), uuid=uuid)
        modsize = 1 + abs(min_value - max_value)
        chunks = r.chunks(length, array_chunk_size)
        queries = []
        arrays = []
        for i in range(chunks[0]):
            chunk = i+1
            queries.append(sql.format(session=r._session(),
                                      uuid=uuid,
                                      chunk=chunk,
                                      mod=modsize,
                                      array_chunk_size=array_chunk_size,
                                      max_block_size=max_block_size,
                                      max_threads=max_threads))
            arrays.append('{session}_{uuid}_{chunk}'.format(session=r._session(),
                                                            uuid=uuid,
                                                            chunk=chunk))
        if chunks[1] > 0:
            arrays.append('{session}_{uuid}_{chunk}'.format(session=r._session(),
                                                            uuid=uuid,
                                                            chunk=len(queries)+1))
            queries.append(sql.format(session=r._session(),
                                      uuid=uuid,
                                      chunk=len(queries)+1,
                                      mod=modsize,
                                      array_chunk_size=chunks[1],
                                      max_block_size=max_block_size,
                                      max_threads=max_threads))
        ctx = vulkn.Vulkn()
        ctx.scheduler.dispatch(queries)
        ctx.session._cache.extend(arrays)
        ctx.session._cache.append(r._ref)
        r._arrays = arrays
        r._rebuild_ref()
        r._length = length
        return r
    
    @classmethod
    @timer
    #def rand(self, min_value, max_value, length, max_block_size=10000, max_threads=2, array_chunk_size=5000000):
    def range(self, start_value, end_value, max_block_size=None, max_threads=2, array_chunk_size=100000000):
        sql = 'CREATE TABLE {session}_{uuid}_{chunk} ENGINE=Memory() AS '
        sql += 'SELECT number AS i, number+({array_chunk_size}*{idx}) AS v FROM numbers_mt({array_chunk_size}) '
        sql += 'SETTINGS max_block_size={max_block_size}, max_threads={max_threads}'
        queries = []
        arrays = []
        uuid = MergeVector.generateUUID()
        max_block_size = max_block_size or array_chunk_size
        r = MergeVector(uuid=uuid,
                        max_block_size=max_block_size,
                        max_threads=max_threads,
                        array_chunk_size=array_chunk_size)
        r._ref = '{session}_{uuid}'.format(session=r._session(), uuid=uuid)
        length = abs(end_value - start_value)
        chunks = r.chunks(length, array_chunk_size)
        for i in range(chunks[0]):
            chunk = i+1
            queries.append(sql.format(session=r._session(),
                                      uuid=uuid,
                                      chunk=chunk,
                                      idx=i,
                                      array_chunk_size=array_chunk_size,
                                      max_block_size=max_block_size,
                                      max_threads=max_threads))
            arrays.append('{session}_{uuid}_{chunk}'.format(session=r._session(),
                                                            uuid=uuid,
                                                            chunk=chunk))
        if chunks[1] > 0:
            arrays.append('{session}_{uuid}_{chunk}'.format(session=r._session(),
                                                            uuid=uuid,
                                                            chunk=len(queries)+1))
            queries.append(sql.format(session=r._session(),
                                      uuid=uuid,
                                      chunk=len(queries)+1,
                                      idx=chunks[0],
                                      array_chunk_size=chunks[1],
                                      max_block_size=max_block_size,
                                      max_threads=max_threads))
        ctx = vulkn.Vulkn()
        ctx.scheduler.dispatch(queries)
        ctx.session._cache.extend(arrays)
        ctx.session._cache.append(r._ref)
        r._arrays = arrays
        r._rebuild_ref()
        r._length = length
        r._sorted = True
        return r

    def agg(self, agg_func):
        return vulkn.Vulkn().q('SELECT {agg_func}(v) FROM {vector}'.format(agg_func=agg_func,
                                                                           vector=self._ref))

    @timer
    def cast(self, to_type):
        sql = 'CREATE TABLE {chunk}_ ENGINE = Memory() AS '
        sql += "SELECT i, cast(v, '{type}') AS v FROM {chunk} "
        sql += 'SETTINGS max_block_size={max_block_size}, max_threads={max_threads}'
        toType = to_type if isinstance(to_type, str) else to_type.CAST
        queries = []
        for a in self._arrays:
            queries.append(sql.format(chunk=a,
                                      type=toType,
                                      max_block_size=self._max_block_size,
                                      max_threads=self._max_threads))
        vulkn.Vulkn().scheduler.dispatch(queries)
        self._rebuild_arrays()
        self._rebuild_ref()
        self._type = toType
        self._sorted = False
        return self

    @timer
    def shuffle(self):
        sql = 'CREATE TABLE {chunk}_ ENGINE = Memory() AS '
        sql += "SELECT rowNumberInBlock() AS i, v FROM {chunk} ORDER BY rand() "
        sql += 'SETTINGS max_block_size={max_block_size}, max_threads={max_threads}'
        queries = []
        for a in self._arrays:
            queries.append(sql.format(chunk=a,
                                      max_block_size=self._max_block_size,
                                      max_threads=self._max_threads))
        vulkn.Vulkn().scheduler.dispatch(queries)
        self._rebuild_arrays()
        self._sorted = False
        return self

    def stats(self):
        sql = 'SELECT min(v) AS min, max(v) AS max, avg(v) AS avg, median(v) AS median FROM {vector}'
        return vulkn.Vulkn().q(sql.format(vector=self._ref))

    def quantiles(self):
        cols = ['quantile({})(v)'.format(str((x+1)/len(self._arrays))) for x in range(len(self._arrays))]
        sql = 'SELECT arrayJoin([{quantiles}]) AS histogram FROM {vector}'
        return vulkn.Vulkn().q(sql.format(quantiles=','.join(cols), vector=self._ref))

    @timer
    def sort(self):
        if self._sorted:
            return self
        if self._type == 'String':
            raise Exception('Supporting for sorting Strings is currently not available in MergeVectors')
        ranges = [h['histogram'] for h in self.quantiles().exec().to_records()]
        queries = []
        for i, a in enumerate(ranges):
            sql = 'CREATE TABLE {chunk}_ ENGINE = Memory() AS '
            sql += 'SELECT rowNumberInBlock() AS i, v FROM (SELECT v FROM {vector} WHERE '
            if i == 0:
                sql += "v < {max_bound} ORDER BY v) "
            elif i == len(self._arrays) - 1:
                sql += "v >= {min_bound} ORDER BY v) "
            else:
                sql += "v >= {min_bound} AND v < {max_bound} ORDER BY v) "
            sql += 'SETTINGS max_block_size={max_block_size}, max_threads={max_threads}'
            queries.append(sql.format(chunk='{chunk}_{idx}'.format(chunk=self._ref, idx=i+1),
                                      max_block_size=round(self._array_chunk_size*1.2),
                                      max_threads=self._max_threads,
                                      vector=self._ref,
                                      max_bound=a,
                                      min_bound=ranges[i-1]))
        vulkn.Vulkn().scheduler.dispatch(queries)
        self._rebuild_arrays()
        self._sorted = True
        return self

    def peek(self):
        sql = 'SELECT * FROM {vector}_1 WHERE i < 20 ORDER BY i LIMIT 20'
        return vulkn.Vulkn().q(sql.format(vector=self._ref))

    @timer
    def take(self, num_rows):
        # TODO: This uses too much memory for large values (1 billion..).
        sql = """
            CREATE TABLE {session}_{new_uuid}_{new_chunk} ENGINE=Memory() AS 
            SELECT
                rowNumberInAllBlocks() AS i, 
                v
            FROM (
                WITH
                    (SELECT groupArray(v) FROM (SELECT v FROM {session}_{uuid}_{chunk} ORDER BY i)) AS `#v`
                SELECT `#v`[(number%length(`#v`))+1] AS v FROM numbers_mt(100000000)
            )"""
        r = MergeVector(max_block_size=self._max_block_size,
                        max_threads=self._max_threads,
                        array_chunk_size=self._array_chunk_size)
        r._length = num_rows
        r._ref = '{session}_{uuid}'.format(session=r._session(), uuid=r._uuid)
        arrays = []
        queries = []
        chunk = 0
        chunks = r.chunks(num_rows, self._array_chunk_size)
        for i in range(chunks[0]):
            new_chunk = i+1
            chunk = (i+1)%len(self._arrays)
            chunk = 1 if chunk == 0 else chunk
            queries.append(sql.format(session=r._session(),
                                      new_uuid=r._uuid,
                                      uuid=self._uuid,
                                      new_chunk=new_chunk,
                                      chunk=chunk,
                                      max_block_size=self._max_block_size,
                                      max_threads=self._max_threads))
            arrays.append('{session}_{uuid}_{chunk}'.format(session=r._session(),
                                                            uuid=r._uuid,
                                                            chunk=new_chunk))
        if (chunks[1] > 0 or 
                (self._array_chunk_size >= self._length and self._array_chunk_size >= num_rows)):
            sql += ' LIMIT {length}'.format(length=chunks[1] if chunks[1] > 0 else num_rows)
            arrays.append('{session}_{uuid}_{chunk}'.format(session=r._session(),
                                                            uuid=r._uuid,
                                                            chunk=len(queries)+1))
            queries.append(sql.format(session=r._session(),
                                      new_uuid=r._uuid,
                                      uuid=self._uuid,
                                      new_chunk=len(queries)+1,
                                      chunk=1 if len(self._arrays) == 1 else (chunk+1)%len(self._arrays),
                                      max_block_size=self._max_block_size,
                                      max_threads=self._max_threads))
        ctx = vulkn.Vulkn()
        ctx.scheduler.dispatch(queries)
        ctx.session._cache.extend(arrays)
        ctx.session._cache.append(r._ref)
        r._arrays = arrays
        r._rebuild_ref()
        return r



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from requests import HTTPSession


from vulkn.clickhouse.client.base import ClickHouseClient


class ClickHouseHTTPClient(ClickHouseClient):
    def __init__(self, config_file: str=None, **kwargs: dict) -> None:
        super(ClickHouseHTTPClient, self).__init__(config_file, **kwargs)      
        self._http = HTTPSession()  

    def _q(self, query: str, settings: dict=None) -> dict:
        r = http.request('get', self._host, self._port)

    def select(self, query: str, settings: dict=None) -> None:
        return self._q(query)


# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import logging


from vulkn.config import FLAGS, SETTINGS, AUTH_OPTS, EXTERNAL_FILE_OPTS
from vulkn.recordset import RecordSet


log = logging.getLogger()


class ClickHouseClient:
    def __init__(self, config_file: str=None, **kwargs: dict) -> None:
        self._config_file = config_file
        self.setFlags(**kwargs)
        self.setOptions(**kwargs)
        self.setAuth(**kwargs)
        self.setExternalFileOptions(**kwargs)
        self._setSettings(**kwargs)

    def setFlags(self, **kwargs: dict) -> any:
        self._setFlags(**kwargs)
        return self

    def setOptions(self, **kwargs: dict) -> any:
        self._setOptions(**kwargs)
        return self

    def setAuth(self, **kwargs: dict) -> any:
        self._setAuth(**kwargs)
        return self

    def setExternalFileOptions(self, **kwargs: dict) -> any:
        self._setExternalFileOptions(**kwargs)
        return self

    def settings(self, **kwargs: dict) -> any:
        self._setSettings(**kwargs)
        return self

    def _q(self, query: str, settings: dict=None) -> any:
        pass

    def execute(self, query: str, settings: dict=None) -> any:
        q = self._q(query, settings=settings)
        return q.returncode

    def select(self, query: str, settings: dict=None) -> RecordSet:
        q = self._q(query, settings=settings)
        return RecordSet(q.stdout)
        
    def insert(self, query: str, settings: dict=None) -> None:
        c = self._cli(settings=settings)

    def insert_blob(self, query: str, input_format: str='TSV', settings: dict=None) -> None:
        c = self._cli(settings=settings)
        c += ['--query', '{} FORMAT {}'.format(query, input_format)]




# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import vulkn.engines
from vulkn.datatable import BaseTableDataTable
from vulkn.utils import timer


class Table:
    def __new__(self, database, table=None):
        import vulkn
        ctx = vulkn.Vulkn()
        if table is None:
            if '.' in database:
                database, table = database.split('.')
            else:
                table = database
                database = None
            database = database or ctx._database
        return BaseTableDataTable(ctx, database, table)

    @classmethod
    @timer
    def fromVector(self, name, columns=(), engine=None, buffer_profile=None, replace=False):
        import vulkn
        database = ''
        table = ''
        ctx = vulkn.Vulkn()
        if isinstance(name, tuple):
            (database, table) = (name[0], name[1])
        if isinstance(name, str):
            (database, table) = name.split('.')
        engine = engine or vulkn.engines.Memory()
        if replace:
            ctx._conn.execute(f'DROP TABLE IF EXISTS {database}.{table}')
        ddl = Table.DDL(database, table, columns, engine, buffer_profile)
        for ddl_query in ddl:
            ctx._conn.execute(ddl_query)
        return BaseTableDataTable(ctx, database, table)

    @staticmethod
    def column_list(columns):
        # TODO: Next release. Whole thing is hacky. Needs refactoring.
        from vulkn.types import ArrayVector, ColumnVector
        from vulkn import Vulkn
        ctx = vulkn.Vulkn()
        rangeCols = []
        selectExprList = []
        q = None
        for idx, column in enumerate(columns):
            if isinstance(column, ArrayVector):
                if hasattr(column, 'vector_name'):
                    selectExprList.append(
                        'v.{index} AS {name}'.format(index=str(idx+1), name=column.vector_name))
                else:
                    selectExprList.append(
                        'v.{index} AS col{index}'.format(index=str(idx+1)))
                rangeCols.append("joinGet('{table}', 'v', 1)".format(table=str(column._cache_table)))
            if isinstance(column, ColumnVector):
                if hasattr(column, 'vector_name'):
                    selectExprList.append(
                        'c{index}.v AS {name}'.format(index=str(idx), name=column.vector_name))
                else:    
                    selectExprList.append(
                        'c{index}.v AS col{index}'.format(index=str(idx)))
                if idx == 0:
                    rangeCols.append('FROM ({}) c0'.format(str(column)))
                else:
                    rangeCols.append(
                        'ANY LEFT JOIN ({subquery}) c{index} ON (c{prevIndex}.i = c{index}.i)'.format(
                            subquery=str(column), index=str(idx), prevIndex=str(idx-1)))
        if isinstance(column, ArrayVector):
            arrayCols = [f'v{i+1}' for i,k in enumerate(rangeCols)]
            ctx.session.optimize(final=True)
            q = """WITH
                    arrayJoin(
                        arrayMap(
                            ({array_cols}) -> ({array_cols}),
                            {source_cols})) AS v
                SELECT {columns}""".format(array_cols=','.join(arrayCols),
                                           source_cols=',\n'.join(rangeCols),
                                           columns=', '.join(selectExprList),
                                           session_cache=ctx.session.session_store)
        if isinstance(column, ColumnVector):
            q = "SELECT {selectExpr} {rangeCols}".format(selectExpr=', '.join(selectExprList),
                                                         rangeCols=' '.join(rangeCols))
        return q

    @staticmethod
    def DDL(database, table, columns, engine, buffer_profile):
        columns = Table.column_list(columns)
        engine = str(engine)
        r = [f'CREATE TABLE {database}.{table} ENGINE = {engine} AS {columns}']
        if buffer_profile:
            r.append(buffer_profile.DDL(database, table))
        return r


table = Table



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import logging


from vulkn.contrib.sqlparse import sqlparse
import vulkn.sql.utils as utils


log = logging.getLogger()
__priority__ = 11


def ASTRewriteChunkByClause(ast):
    ret = sqlparse.parse(str(list(_rewrite(ast))[0]))[0]
    return ret

def _rewrite(ast):
    tree = ast.tokens if isinstance(ast, sqlparse.sql.TokenList) else ast
    for idx, token in enumerate(tree):
        if token.ttype == sqlparse.tokens.Keyword and token.value.upper().startswith('CHUNK'):
            yield _ASTRewriteChunkByClause(ast, ast, idx).rewrite()
            break
    else:
        yield ast

class _ASTRewriteChunkByClause():
    def __init__(self, tokens, ast, curr_idx):
        self._tokens = tokens.tokens
        self._ast = ast
        self._curr_idx = curr_idx

    def rewrite(self):
        args = utils.get_keyword_function_args(self._ast, self._curr_idx)
        (key, chunk_size) = (args[0], args[1])
        utils.del_keyword_function_node(self._tokens, self._ast, self._curr_idx)
        for j in range(0, len(self._ast.tokens)):
            if isinstance(self._ast[j], sqlparse.sql.Where):
                self._ast[j].tokens.append(sql.Token('Literal', ' AND ({chunk_clause})'))
                break
            if self._ast[j].ttype == sqlparse.tokens.Keyword and (
                    self._ast[j].value.upper().startswith('GROUP') or
                    self._ast[j].value.upper().startswith('ORDER') or
                    self._ast[j].value.upper().startswith('LIMIT')):
                self._ast.insert_before(j - 2, sqlparse.sql.Token('Literal', ' WHERE ({chunk_clause})'))
                break
        else:
            self._ast.tokens.append(sqlparse.sql.Token('Literal', ' WHERE AND ({chunk_clause})'))
        chunk_clause = f'cityHash64({key})%{chunk_size}={{chunk_step}}'
        l = [
            ''.join(map(str, self._ast)).format(chunk_clause=chunk_clause.format(chunk_step=k))
            for k in range(0, int(chunk_size))]
        q = '\nUNION ALL\n'.join(s.strip() for s in l)
        return q



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import TypeBase, func
from vulkn.types.array import Array
from vulkn.types.integer import UInt8, UInt32, UInt64
from vulkn.utils import item_index


class String(TypeBase):
    TYPE = 'String'
    CAST = 'String'
    _METHODS = {
        'alphaTokens': 'Array',
        'empty': 'UInt8',
        'notEmpty': 'UInt8',
        'length': 'UInt64',
        'lengthUTF8': 'UInt64',
        'char_length': 'UInt64',
        'CHAR_LENGTH': 'UInt64',
        'character_length': 'UInt64',
        'CHARACTER_LENGTH': 'UInt64',
        'lower': None,
        'lcase': None,
        'upper': None,
        'ucase': None,
        'lowerUTF8': None,
        'upperUTF8': None,
        'isValidUTF8': 'UInt8',
        'toValidUTF8': None,
        'reverse': None,
        'reverseUTF8': None,
        'base64Encode': None,
        'base64Decode': None,
        'tryBase64Decode': None,
        'unhex': None,
        'trimLeft': None,
        'trimRight': None,
        'trimBoth': None,
        'CRC32': 'UInt32',
        'CRC64': 'UInt64',
        'CRC32IEEE': 'UInt32',
        'MD5': None,
        'SHA1': None,
        'SHA224': None,
        'SHA256': None,
        'basename': None
    }

    def __add__(self, right):
        return self.concat(right)

    def __getitem__(self, index):
        start, stop = item_index(index)
        if stop is None:
            stop = 1
        return String(func('substring', self._value, start, stop))
        
    def substring(self, offset, length=None):
        if length:
            return String(func('substring', self._value, offset, length))
        return String(func('substring', self._value, offset))

    mid = substring
    substr = substring

    def len(self):
        return UInt64(func('length', self._value))

    def substringUTF8(self, offset, length=None):
        if length:
            return String(func('substringUTF8', self._value, offset, length))
        return String(func('substringUTF8', self._value, offset))

    def trim(self, trim_str: str='\\\s*'):
        return self.replaceRegexpAll('^{}|{}$'.format(trim_str, trim_str), '')

    strip = trim
        
    def isnumeric(self):
        return self.match('^[0-9]*$')

    def isdecimal(self):
        return self.match('^[0-9]*\.[0-9]*$')

    def ltrim(self, trim_str: str='\\\s*'):
        return self.replaceRegexpOne('^{}'.format(trim_str), '')

    lstrip = ltrim
        
    def rtrim(self, trim_str: str='\\\s*'):
        return self.replaceRegexpOne('{}$'.format(trim_str), '')

    rstrip = rtrim
            
    def position(self, needle: str):
        return String(func('position', self._value, needle))

    locate = position

    def positionUTF8(self, needle: str):
        return String(func('positionUTF8', self._value, needle)) 

    def positionCaseInsensitive(self, needle: str):
        return String(func('positionCaseInsensitive', self._value, needle))

    def positionCaseInsensitiveUTF8(self, needle: str):
        return String(func('positionCaseInsensitiveUTF8', self._value, needle))

    def replaceOne(self, pattern, replacement):
        return String(func('replaceOne', self._value, pattern, replacement))

    def replaceAll(self, pattern, replacement):
        return String(func('replaceAll', self._value, pattern, replacement))

    def replaceRegexpOne(self, pattern, replacement):
        return String(func('replaceRegexpOne', self._value, pattern, replacement))

    def replaceRegexpAll(self, pattern, replacement):
        return String(func('replaceRegexpAll', self._value, pattern, replacement))

    def replace(self, pattern, replacement, count=None):
        if count:
            return self.replaceOne(pattern, replacement)
        else:
            return self.replaceAll(pattern, replacement)

    def match(self, pattern: str):
        return String(func('match', self._value, pattern))

    def extract(self, pattern: str):
        return String(func('extract', self._value, pattern))

    def extractAll(self, pattern: str):
        return String(func('extractAll', self._value, pattern))

    def like(self, pattern: str):
        return UInt8(func('like', self._value, pattern))

    def notLike(self, pattern: str):
        return String(func('notLike', self._value, pattern))

    def splitByString(self, separator_arg: str) -> Array:
        return Array(func('splitByString', separator_arg, self._value))

    def splitByChar(self, separator_arg: str) -> Array:
        return Array(func('splitByChar', separator_arg, self._value))

    def concat(self, *args):
        return String(func('concat', *([self._value] + [*args])))

    def concatAssumeInjective(self, *args):
        return String(func('concatAssumeInjective', *([self._value] + [*args])))
        
    def appendTrailingCharIfAbsent(self, char_arg: str):
        return String(
            func('appendTrailingCharIfAbsent', self._value, char_arg))

    append_if_missing = appendTrailingCharIfAbsent

    def convertCharset(self, from_arg: str, to_arg: str):
        return String(func('convertCharset', self._value, from_arg, to_arg))

    def splitlines(self) -> Array:
        return self.splitByChar('\\n')

    def startswith(self, pattern: str):
        return self.like('{}%'.format(pattern))
    
    def endswith(self, pattern: str):
        return self.like('%{}'.format(pattern))

    split = splitByString

    def join(self, arr):
        return String(func('arrayStringConcat', self._value, arr))



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


from vulkn.types.base import func, Literal
from vulkn.types.numeric import Numeric


class Integer(Numeric):
    _METHODS = {
        'abs': 'Integer',
        'negate': 'Integer',
        'sum': None}

    def plus(self, right):
        return Integer(func('plus', self._value, right))

    def minus(self, right):
        return Integer(func('minus', self._value, right))

    def multiply(self, right):
        return Integer(func('multiply', self._value, right))

    def divide(self, right):
        return Integer(func('divide', self._value, right))

    def intDiv(self, right):
        return Integer(func('intDiv', self._value, right))

    def intDivOrZero(self, right):
        return Integer(func('intDivOrZero', self._value, right))

    def modulo(self, right):
        return Integer(func('modulo', self._value, right))

    def gcd(self, right):
        return Integer(func('gcd', self._value, right))

    def lcm(self, right):
        return Integer(func('lcm', self._value, right))

    def vectorAccumulate(self):
        return Integer(func('vectorAccumulate', Literal('__{}'.format(str(self._value)[1:-1]))))


class Int8(Integer):
    MIN = -128
    MAX = 127
    TYPE = 'Int8'
    CAST = 'Int8'


class Int16(Integer):
    MIN = -32768
    MAX = 32767
    TYPE = 'Int16'
    CAST = 'Int16'


class Int32(Integer):
    MIN = -2147483648
    MAX = 2147483647
    TYPE = 'Int32'
    CAST = 'Int32'


class Int64(Integer):
    MIN = -9223372036854775808
    MAX = 9223372036854775807
    TYPE = 'Int64'
    CAST = 'Int64'


class UInt8(Integer):
    MIN = 0
    MAX = 255
    TYPE = 'UInt8'
    CAST = 'UInt8'


class UInt16(Integer):
    MIN = 0
    MAX = 65535
    TYPE = 'UInt16'
    CAST = 'UInt16'


class UInt32(Integer):
    TYPE = 'UInt32'
    CAST = 'UInt32'
    _METHODS = {
        'IPv4NumToString': 'String',
        'IPv4NumToStringClassC': 'String',
        'IPv6NumToString': 'String',
        'rand': None,
        'randConstant': None}
    MIN = 0
    MAX = 4294967295


class UInt64(Integer):
    TYPE = 'UInt64'
    CAST = 'UInt64'
    _METHODS = {
        'rand64': None,
        'intExp2': None,
        'intExp10': None,
        'sum': None}
    MIN = 0
    MAX = 18446744073709551615



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


import ast
import logging


from vulkn.clickhouse import sqlformat
from vulkn.utils import timer


log = logging.getLogger()


class MutationManager:
    def __init__(self, ctx):
        self._ctx = ctx

    def _list(self):
        return self._ctx.q('SELECT * FROM system.mutations').exec()

    def show(self):
        self._list().show()

    def to_records(self):
        return self._list().to_records()

    def failed(self):
        return self._ctx.q("SELECT * FROM system.mutations WHERE latest_fail_reason != ''").exec()

    def kill(self, mutation_id):
        r = self._ctx.q(f"SELECT count() AS count FROM system.mutations WHERE mutation_id = '{mutation_id}'").exec().to_records()
        if r[0]['count'] == 0:
            raise Exception(f'Mutation {mutation_id} does not exist')
        return self._ctx.exec(f"KILL MUTATION WHERE mutation_id = '{mutation_id}'".format(mutation_id))

    def kill_all(self):
        return self._ctx.exec('KILL MUTATION WHERE 1 = 1')
    
    def active(self):
        return self._ctx.q("SELECT * FROM system.mutations WHERE is_done = 0").exec()


class VulknClickHouseDatabaseMixIn:
    def _set_db(self, table):
        if '.' not in table:
            table = '{database}.{table}'.format(database=self._database, table=table)
        return table

    def use(self, database):
        self._database = database
        self._conn = self._conn.setAuth(host=self._host,
                                        port=self._port,
                                        user=self._user,
                                        password=self._password,
                                        database=self._database)

    def desc(self, table):
        return self.q('DESC {}'.format(self._set_db(table))).exec()

    d = desc

    def get_create(self, table):
        return sqlformat(
            self.q('SHOW CREATE {}'.format(self._set_db(table))).e().to_records()[0]['statement'],
            oneline=False)

    getCreate = get_create

    def show_create(self, table):
        print(self.getCreate(table))

    showCreate = show_create

    c = showCreate

    def profile(self, table):
        def profile_column(table, column):
            column_type = column._col_type
            column_name = column._name
            profile_data = {}
            histogram = None
            profiles = []
            profile_data['uniq_values'] = (self.q(
                f'SELECT uniqCombined("{column_name}") AS uniq_values FROM {table}')
                .exec()
                .to_records())[0]['uniq_values']
            if profile_data['uniq_values'] <= 10:
                histogram = f"""
                    (SELECT
                        groupArray(({{subject}}, count))
                    FROM (
                        SELECT
                            {{subject}},
                            count() AS count
                        FROM {table}
                        GROUP BY {{subject}}
                        ORDER BY {{subject}}))"""
            else:
                histogram = f"""arrayMap(
                    x -> ({{subject_from}}(x.1), {{subject_from}}(x.2), round(x.3)),
                    histogram(10)({{subject_to}}))"""
            available_profiles = [
                "min({subject}) AS min",
                "max({subject}) AS max", 
                "mode",
                "countIf(isNull({subject})) AS nulls", 
                "count() AS total_records", 
                "toTypeName(any({subject})) AS data_type",
                "varSamp({subject_to}) AS var_samp",
                "varPop({subject_to}) AS var_pop",
                "stddevSamp({subject_to}) AS stddev_samp",
                "stddevPop({subject_to}) AS stddev_pop",
                "skewPop({subject_to}) AS skew_pop",
                "skewSamp({subject_to}) AS skew_samp",
                "kurtPop({subject_to}) AS kurt_pop",
                "kurtSamp({subject_to}) AS kurt_samp"
            ]
            if 'Date' in column_type:
                profiles.append(
                    "toDateTime(round(avg(toUInt32({subject})))) AS mean".format(subject=column_name))
                profiles.append(f"{histogram} AS histogram".format(subject=column_name,
                                                                   subject_to=f'toUInt32({column_name})',
                                                                   subject_from=f'toDateTime'))
                profiles.append("median({subject}) AS median".format(subject=column_name))
                for p in available_profiles:
                    profiles.append(p.format(subject=column_name,
                                             subject_to=f'toUInt32({column_name})',
                                             subject_from=f'toDateTime'))
            elif 'Int' in column_type:
                profiles.append(
                    "round(avg({subject})) AS mean".format(subject=column_name))
                profiles.append(f"{histogram} AS histogram".format(subject=column_name,
                                                                   subject_to=column_name,
                                                                   subject_from=''))
                profiles.append("median({subject}) AS median".format(subject=column_name))
                for p in available_profiles:
                    profiles.append(p.format(subject=column_name,
                                             subject_to=column_name,
                                             subject_from=f'round'))
            elif 'String' in column_type:
                profiles.append(
                    "round(avg(length({subject}))) AS mean_length".format(subject=column_name))
                profiles.append(
                    "round(min(length({subject}))) AS min_length".format(subject=column_name))
                profiles.append(
                    "round(max(length({subject}))) AS max_length".format(subject=column_name))
                profiles.append(
                    "round(median(length({subject}))) AS median_length".format(subject=column_name))
                for p in available_profiles:
                    profiles.append(p.format(subject=column_name,
                                             subject_to=f'length({column_name})',
                                             subject_from=f'round'))
            else:
                profiles.append("avg({subject}) AS mean".format(subject=column_name))
                profiles.append(f"{histogram} AS histogram".format(subject=column_name,
                                                                   subject_to=column_name,
                                                                   subject_from=''))
                profiles.append("median({subject}) AS median".format(subject=column_name))
                for p in available_profiles:
                    profiles.append(p.format(subject=column_name,
                                             subject_to=column_name,
                                             subject_from=''))
            profile_query = f"""
                WITH
                    (SELECT topK(10)({column_name}) FROM {table}) AS topK,
                    topK[1] AS mode
                SELECT
                    topK,
                    {','.join(profiles)}
                FROM {table}"""
            
            profile_data.update(self.q(profile_query).exec().to_records()[0])
            if 'histogram' in profile_data:
                profile_data['histogram'] = ast.literal_eval(profile_data['histogram'])
            if 'topK' in profile_data:
                profile_data['topK'] = ast.literal_eval(profile_data['topK'])
            return profile_data

        schema = self.getSchema(table)
        profiles = {}
        for column in schema:
            profiles[column._name] = profile_column(table, column)
        return profiles

    def get_schema(self, database=None, table=None):
        from vulkn.types import ColumnType
        schema = []
        if table is None and database is not None:
            table = database
            database = None
        if table is None:
            raise Exception('No table specified')
        if database is None:
            database = self._database
        sql = """
            SELECT
                name,
                type,
                default_kind,
                default_expression,
                compression_codec
            FROM system.columns 
            WHERE database = {database:String} AND table = {table:String}"""
        columns = self.q(sql).exec(database=database, table=table).to_records()
        for r in columns:
            schema.append(
                ColumnType(r['name'],
                          r['type'],
                          default_kind=r['default_kind'],
                          default_expression=r['default_expression'],
                          compression_codec=r['compression_codec']))
        return schema

    getSchema = get_schema
    schema = get_schema

    def db(self):
        return self.q('SHOW DATABASES').exec()

    def tables(self, database=None):
        if database is None:
            database = self._database
        return self.q("""
            SELECT database, name, engine, partition_key, sorting_key
            FROM system.tables
            WHERE database = {database:String}
            ORDER BY database != 'system' DESC, database,  engine, name""").exec(database=database)

    t = tables
    
    def optimize(self, database, table, partition=None, final=False, deduplicate=False):
        ddl = f'OPTIMIZE TABLE {database}.{table}'
        if final:
            ddl += ' FINAL'
        if deduplicate:
            ddl += ' DEDUPLICATE'
        return self.exec(ddl)

    def parts(self, database=None, table=None):
        if table is None and database is not None:
            table = database
            database = None
        if table is None:
            raise Exception('No table specified')
        if database is None:
            database = self._database
        return self.q("""
            SELECT partition, rows, path
            FROM system.parts
            WHERE
                active
                AND (
                    (database = {database:String} AND table = {table:String}) 
                    OR database || '.' || table = {table:String})
            ORDER BY partition""").exec(database=database, table=table)

    def drop_part(self, database=None, table=None, partitions=None):
        if partitions is None and database is not None and table is not None:
            partitions = table
            table = database
            database = self._database
        if ((partitions is None or database is None or table is None) 
                or not isinstance(partitions, list)):
            raise Exception('Invalid parameters passed to dropPart')
        ddl = f'ALTER TABLE {database}.{table} DROP PARTITION '
        for p in partitions:
            if not self.exec("{} '{}'".format(ddl, p)):
                raise Exception(f'Unable to drop partition {p}')
        return True

    dropPart = drop_part

    def clone_table(self, from_database, from_table, database, table):
        ddl = f'CREATE TABLE {database}.{table} AS {from_database}.{from_table}'
        return self.exec(ddl)

    cloneTable = clone_table

    def copy_table(self, from_database, from_table, database, table):
        if not self.cloneTable(from_database, from_table, database, table):
            raise Exception('Unable to clone table')
        return self.exec(f'INSERT INTO {database}.{table} SELECT * FROM {from_database}.{from_table}')

    copyTable = copy_table

    def create_table(self, database=None, table=None, schema=None, engine=None, exists_ok=False):
        schema_ddl = ', '.join(['{} {}'.format(k, v) for k, v in schema.items()])
        ddl = f'CREATE TABLE {database}.{table} ({schema_ddl}) ENGINE = {engine}'
        return self.exec(ddl)

    createTable = create_table

    def create_sink(self, database, sink, schema, exists_ok=False):
        schema_ddl = ', '.join(['{} {}'.format(k, v) for k, v in schema.items()])
        ddl = f'CREATE TABLE {database}.{sink} ({schema_ddl}) ENGINE = Null'
        return self.exec(ddl)

    createSink = create_sink

    def create_view(self, database, view, query, exists_ok=False):
        ddl = f'CREATE VIEW {database}.{view} AS {query}'
        return self.exec(ddl)

    createView = create_view

    def create_mat_view(self,
                      database,
                      view,
                      query,
                      to_database=None,
                      to_table=None,
                      engine=None,
                      exists_ok=False):
        ddl = f'CREATE MATERIALIZED VIEW {database}.{view} '
        if to_database is not None and to_table is not None:
            ddl = f'TO {to_database}.{to_table} AS {query}'
        else:
            engine = engine or engines.MergeTree()
            ddl = f'ENGINE = {engine} AS {query}'
        return self.exec(ddl)

    createMatView = create_mat_view

    def drop_table(self, database, table):
        ddl = f'DROP TABLE IF EXISTS {database}.{table}'
        return self.exec(ddl)

    dropTable = drop_table

    def drop_columns(self, database, table, columns):
        r = True
        for c in columns:
            try:
                if not self.exec(f'ALTER TABLE {database}.{table} DROP COLUMN {c}'):
                    r = False
                    log.info(f'Unable to drop column {c} on {database}.{table}')
            except:
                r = False
                log.info(f'Unable to drop column {c} on {database}.{table}')
        return r

    def truncate_table(self, database, table):
        ddl = f'TRUNCATE TABLE {database}.{table}'
        return self.exec(ddl)

    truncateTable = truncate_table

    def table_exists(self, database, table):
        exists = f"""
            SELECT count()>0 AS exists 
            FROM system.tables
            WHERE (name = '{table}') OR concat(database, '.', name) = '{database}.{table}'"""
        return self.q(exists).exec().to_records()[0]['exists'] == 1

    tableExists = table_exists

    dropSink = drop_table
    drop_sink = drop_table
    dropView = drop_table
    drop_view = drop_table
    dropMatView = drop_table
    drop_mat_view = drop_table

    @timer
    def exec(self, query):
        return self.scheduler.dispatch(query)

    e = exec


class VulknSystemManager:
    def __init__(self, ctx):
        self._ctx = ctx
        self.exec = self._ctx.exec

    def reload_dicts(self):
        return self.exec('SYSTEM RELOAD DICTIONARIES')

    def reload_dict(self, dictionary):
        return self.exec(f'SYSTEM RELOAD DICTIONARY {dictionary}')

    def drop_dns_cache(self):
        return self.exec('SYSTEM DROP DNS CACHE')

    def drop_mark_cache(self):
        return self.exec('SYSTEM DROP MARK CACHE')

    def flush_logs(self):
        return self.exec('SYSTEM FLUSH LOGS')

    def reload_config(self):
        return self.exec('SYSTEM RELOAD CONFIG')

    def shutdown(self):
        return self.exec('SYSTEM SHUTDOWN')

    def kill(self):
        return self.exec('SYSTEM KILL')

    def stop_dist_sends(self, database, table):
        return self.exec(f'SYSTEM STOP DISTRIBUTED SENDS {database}.{table}')

    def flush_dist(self, database, table):
        return self.exec(f'SYSTEM FLUSH {database}.{table}')

    def start_dist_sends(self, database, table):
        return self.exec(f'SYSTEM START DISTRIBUTED SENDS {database}.{table}')

    def stop_merges(self, database, table):
        return self.exec(f'SYSTEM STOP MERGES {database}.{table}')

    def start_merges(self, database, table):
        return self.exec(f'SYSTEM START MERGES {database}.{table}')



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Jason Godden <jason@godden.id.au>
# Copyright (c) 2019, VulknData Pty Ltd
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-only


class VulknCatalogViewer:
    def __init__(self, ctx, catalog='system', hidden=None):
        self._ctx = ctx
        self._catalog = catalog
        self._hidden = hidden or []

    def __getattr__(self, attrname):
        if attrname in self._hidden:
            raise AttributeError
        sql = "SELECT name FROM system.tables WHERE name = {name:String} AND database = {catalog:String}"
        t = self._ctx.q(sql).exec(catalog=self._catalog, name=attrname).to_records()
        if len(t) > 0:
            return self._ctx.q(f'SELECT * FROM "{self._catalog}"."{attrname}"').exec()
        else:
            raise AttributeError

    def help(self):
        sql = "SELECT name FROM system.tables WHERE database = {catalog:String} ORDER BY name"
        return [n['name'] for n in self._ctx.q(sql).exec(catalog=self._catalog).to_records()]


class VulknCatalog:
    def __init__(self, ctx):
        self._ctx = ctx

    @property
    def h(self):
        print('\n'.join([
            'tables', 'table_engines', 'mutations']))

    @property
    def tables(self):
        return self._ctx.q("""
            SELECT
                database, name, engine, partition_key, sorting_key
            FROM system.tables
            ORDER BY 
                database != 'system' DESC,
                database, 
                engine,
                name""").s

    @property
    def table_engines(self):
        return self._ctx.q("""
            SELECT
                name
            FROM system.table_engines
            ORDER BY
                name LIKE '%MergeTree%',
                name LIKE '%Replicated%', 
                name""").s

    @property
    def mutations(self):
        return self._ctx.q("""
            select 
                create_time, database || '.' || table as table, command, parts_to_do, is_done, 
                latest_failed_part, latest_fail_time, latest_fail_reason
            from system.mutations order by table""").s

    @property
    def databases(self):
        return self._ctx.q("select * from system.databases").s

    @property
    def settings(self):
        return self._ctx.q("SELECT name, value FROM system.settings ORDER BY name").s

    @property
    def functions(self):
        return self._ctx.q("select name from system.functions where is_aggregate = 0 order by name").s

    @property
    def aggfunctions(self):
        return self._ctx.q("select name from system.functions where is_aggregate = 1 order by name").s

    @property
    def clusters(self):
        return self._ctx.q('select * from system.clusters order by cluster').s

    @property
    def formats(self):
        return self._ctx.q('select * from system.formats order by name').s

    @property
    def parts(self):
        return self._ctx.q('select * from system.parts order by database, table').s

    @property
    def metrics(self):
        return self._ctx.q('select * from system.metrics order by metric').s

    @property
    def merge_tree_settings(self):
        return self._ctx.q('select * from system.merge_tree_settings order by name').s

    @property
    def events(self):
        return self._ctx.q('select * from system.events order by event').s

    @property
    def dicts(self):
        return self._ctx.q('select * from system.dictionaries').s

    @property
    def datatypes(self):
        return self._ctx.q('select * from system.data_type_families').s

    @property
    def build_options(self):
        return self._ctx.q('select * from system.build_options').s

    @property
    def columns(self):
        return self._ctx.q('select * from system.columns order by database, table, name').s